package main

import (
	"bytes"
	"crypto/cipher"
	"crypto/rand"
	"encoding/hex"
	"fmt"
	"os"

	"github.com/aead/camellia"
	"github.com/d1str0/pkcs7"
)

const BLOCKSIZE int = 16
const TAGLEN int = BLOCKSIZE*8/2 - 1
const MENU string = `====================
1. Help
2. Generate MAC
3. Verify
4. Exit
====================
`

func XOR(a, b []byte) []byte {
	if len(a) > len(b) {
		b = append(b, make([]byte, len(a)-len(b))...)
	} else {
		a = append(a, make([]byte, len(b)-len(a))...)
	}
	c := make([]byte, len(a))
	for i := 0; i < len(a); i++ {
		c[i] = a[i] ^ b[i]
	}
	return c
}

func GenerateRandBytes(n int) []byte {
	b := make([]byte, n)
	rand.Read(b)
	return b
}

func BitStringToBytes(s string) []byte {
	b := make([]byte, (len(s)+(8-1))/8)
	for i, r := range s {
		b[i>>3] |= byte(r-'0') << uint(7-i&7)
	}
	return b
}

func Convert(data []byte, msb bool) []byte {
	var b bytes.Buffer
	for _, v := range data {
		b.WriteString(fmt.Sprintf("%08b", v))
	}
	var substr string
	if msb {
		substr = b.String()[:TAGLEN]
	} else {
		substr = b.String()[len(b.String())-TAGLEN:]
	}
	substr = "0" + substr
	return BitStringToBytes(substr)
}

type RCBC struct {
	key    []byte
	cipher cipher.Block
}

func NewRCBC() *RCBC {
	key := GenerateRandBytes(BLOCKSIZE)
	cipher, err := camellia.NewCipher(key)
	if err != nil {
		// Exit with error message
		fmt.Println("Error.")
		os.Exit(1)
	}
	return &RCBC{key, cipher}
}

func (r *RCBC) GenerateMAC(data []byte, n int) []byte {
	var bcnt int
	if len(data)%n == 0 {
		bcnt = len(data) / n
	} else {
		bcnt = len(data)/n + 1
	}
	blocks := make([][]byte, bcnt)
	for i := 0; i < len(data); i += n {
		mx := i + n
		if mx > len(data) {
			mx = len(data)
		}
		blocks[i/n] = data[i:mx]
	}

	y := make([]byte, n)
	for i := 0; i < len(blocks)-1; i++ {
		x := XOR(blocks[i], y)
		r.cipher.Encrypt(y, x)
	}

	if len(blocks[len(blocks)-1]) == n {
		x := XOR(blocks[len(blocks)-1], y)
		r.cipher.Encrypt(y, x)
		return Convert(y, true)
	} else {
		var padded []byte
		padded, err := pkcs7.Pad(blocks[len(blocks)-1], n)
		if err != nil {
			fmt.Println("Error.")
			os.Exit(1)
		}
		x := XOR(padded, y)
		r.cipher.Encrypt(y, x)
		return Convert(y, false)
	}
}

func (r *RCBC) VerifyMAC(data []byte, n int, tag []byte) bool {
	return string(r.GenerateMAC(data, n)) == string(tag)
}

func main() {
	var flag string
	flag = "SEKAI{test_flag}"

	r := NewRCBC()
	counter := 3
	history_msg := make([][]byte, 0)

	var choice int
	var msg, tag string

	for {
		fmt.Print(MENU)
		fmt.Print("Enter your choice: ")
		fmt.Scan(&choice)

		switch choice {
		case 1:
			fmt.Println("RCBC is a secure MAC that is robust to all attacks.")
			fmt.Println("You can make no more than three queries. Can you forge a valid MAC?")
		case 2:
			if counter == 0 {
				fmt.Println("You cannot make any more queries.")
			} else {
				counter -= 1
				fmt.Print("Enter message in hex: ")
				fmt.Scan(&msg)
				decodedBytes, err := hex.DecodeString(msg)
				if err != nil {
					fmt.Println("Error.")
					os.Exit(1)
				}
				if !bytes.Contains(decodedBytes, []byte("Sekai")) {
					fmt.Println("Sorry.")
				} else {
					history_msg = append(history_msg, decodedBytes)
					fmt.Println("MAC:", hex.EncodeToString(r.GenerateMAC(decodedBytes, BLOCKSIZE)))
				}
			}
		case 3:
			fmt.Print("Enter message in hex: ")
			fmt.Scan(&msg)
			decodedBytes, err := hex.DecodeString(msg)
			if err != nil {
				fmt.Println("Error.")
				os.Exit(1)
			}

			for _, m := range history_msg {
				if bytes.Equal(decodedBytes, m) {
					fmt.Println("Sorry.")
					os.Exit(0)
				}
			}

			fmt.Print("Enter MAC in hex: ")
			fmt.Scan(&tag)
			decodedTag, err := hex.DecodeString(tag)
			if err != nil {
				fmt.Println("Error.")
				os.Exit(1)
			}

			if r.VerifyMAC(decodedBytes, BLOCKSIZE, decodedTag) {
				fmt.Println("Hmmm the scheme seems broken. Here is your flag:", flag)
			} else {
				fmt.Println("Verification failed.")
			}
			os.Exit(0)
		case 4:
			fmt.Println("Bye!")
			os.Exit(0)
		default:
			os.Exit(0)
		}
	}
}
